
# install_packages3.py - Language Support Installtion using PackageKit
#
# Copyright (C) 2013-2015 Red Hat, Inc.
# Authors: Parag Nemade <pnemade@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function

import dnf
import subprocess
import os
from gi.repository import Gio, GLib

import gettext
_ = lambda x: gettext.translation(
    "system-config-language", fallback=True).gettext(x) if x != "" else ""
N_ = lambda x: x


class CompsParser(object):

    def __init__(self):
        self.__cached_c_element_tree = None

    def _c_element_tree_import(self):
        """ Importing xElementTree all the time, when we often don't need it, is
            a huge timesink. This makes python -c 'import yum' suck. So we hide
            it behind this function. And have accessors. """

        if self.__cached_c_element_tree is None:
            from xml.etree import cElementTree
            self.__cached_c_element_tree = cElementTree

    def c_elementtree_iterparse(self, filename):
        """ Lazily load/run: cElementTree.iterparse """
        self._c_element_tree_import()
        return self.__cached_c_element_tree.iterparse(filename)

    def iterparse(self, filename):
        try:
            for elem in self.c_elementtree_iterparse(filename):
                yield elem
        except SyntaxError as elem:
            print('Syntax error in file %s for %s' % (filename, elem))


class InstallLangpack(object):

    """ Class for installing packages using PackageKit """

    def __init__(self):
        """ init method """
        self.conditional_pkgs = {}
        self.base = dnf.Base()
        self.list_to_install = []
        self.uniq_installed_pkgs = []
        self.langlist_with_cc = ['zh_CN', 'zh_TW', 'pt_BR']
        self.conffile = '/var/lib/dnf/plugins/langpacks/installed_langpacks'
        # we are not sure if conffile already exists on the system or
        # user moved or deleted it. To make sure we have conffile before
        # using any langpacks command let's try to create it first.
        self.conffile_dir = os.path.dirname(self.conffile)
        if not os.path.exists(self.conffile_dir):
            try:
                os.makedirs(self.conffile_dir)
            except (IOError, OSError) as fperror:
                print('Unable to create file : %s' % self.conffile)
                return
            try:
                with open(self.conffile, 'a'):
                    pass
            except IOError:
                print("Unable to create installed_langpacks file")

    def write_installed_langpacks(self, langplist):
        """ Write the installed langpacks from conffile """

        if not self.conffile:
            return
        confdir = os.path.dirname(self.conffile)
        if not os.path.exists(confdir):
            try:
                os.makedirs(confdir, 0o755)
            except (IOError, OSError) as e:
                return
        try:
            tmp = open(self.conffile + ".tmp", "w+")
            for line in langplist:
                tmp.write(line + "\n")
            tmp.close()
            os.rename(self.conffile + ".tmp", self.conffile)
        except (IOError, OSError) as e:
            return

    def read_installed_langpacks(self):
        """ Read the installed langpacks from conffile """

        if not self.conffile:
            return []
        ret = []
        try:
            f = open(self.conffile, "r")
            flist = f.readlines()
            f.close()
        except (IOError, OSError) as e:
            return []
        for item in flist:
            item = item.strip()
            ret.append(item)
        return ret

    def add_langpack_to_installed_list(self, lang):
        """ Add the language code to the installed languages file """
        modified = 0
        langplist = self.read_installed_langpacks()
        if lang not in langplist:
            langplist.append(lang)
            modified = 1
        if modified:
            self.write_installed_langpacks(langplist)

    def check_if_already_installed(self, check_lang):
        """ Check if requested langpack is already installed on the system """
        langpack_list = self.read_installed_langpacks()
        if check_lang not in langpack_list:
            print(("{0} language not installed").format(check_lang))
            return "no"
        else:
            print(("{0} language is already installed").format(check_lang))
            return "yes"

    @classmethod
    def get_language_packages(cls, lang):
        """ Find package list required to install the requested langpack """
        lang_pkgs = []
        cmd = "dnf langinfo {0}".format(lang)
        cmdoutput = subprocess.check_output(cmd, shell=True)
        for line in cmdoutput.decode('UTF-8').splitlines():
            if line.startswith("  "):
                lang_pkgs.append(line[2:])
        print("Langpacks provides packages => {0}".format(lang_pkgs))

        return lang_pkgs

    def setup_conditional_pkgs(self, repos):
        """ This is called to check if cond pkg already setup """
        if not self.conditional_pkgs:
            for repo in repos:
                if not repo.enablegroups:
                    continue
                if not repo.metadata:
                    continue
                comps_fn = repo.metadata.comps_fn
                if comps_fn is None:
                    continue

                if repo.md_only_cached:
                    infile = dnf.yum.misc.calculate_repo_gen_dest(comps_fn,
                                                                  'groups.xml')
                    if not os.path.exists(infile):
                        # root privileges are needed for comps decompression
                        continue
                else:
                    infile = dnf.yum.misc.repo_gen_decompress(
                        comps_fn, 'groups.xml')

                comparse = CompsParser()
                for tp in comparse.iterparse(infile):
                    elem = tp[1]
                    if elem.tag == "langpacks":
                        for child in elem.getchildren():
                            if child.tag != "match":
                                continue
                            name = child.get("name")
                            install = child.get("install")

                            if name not in self.conditional_pkgs:
                                self.conditional_pkgs[name] = []
                            self.conditional_pkgs[name].append(install)

    @classmethod
    def get_matches(cls, availpkg, llist):
        ret = []
        for match in llist:
            try:
                p = availpkg.filter(provides=match)
                if p[0].name:
                    ret.append(p[0].name)
            except:
                pass
        return ret

    def find_matching_pkgs(self, ipkgs, lang):
        """ Find all conditional language packages """
        pkgmatches = []
        for basepkg in self.conditional_pkgs:
            if basepkg in ipkgs:
                conds = self.conditional_pkgs[basepkg]
                patterns = [x % (lang,) for x in conds]
                shortlang = lang.split('_')[0]
                if shortlang != lang:
                    patterns = patterns + [x % (shortlang,) for x in conds]
                for p in patterns:
                    if p not in pkgmatches:
                        # just pattern matched pkgs irrespective of its
                        # existence
                        pkgmatches.append(p)
        return pkgmatches

    def install_packages3(self, lang):
        """ Find which packages are required for installation """
        pkgmatches = []
        ipkgs = []
        self.base.read_all_repos()
        repos = self.base.repos.iter_enabled()
        self.base.fill_sack()
        self.setup_conditional_pkgs(repos)
        allpkg = self.base.sack.query()
        instpkg = allpkg.installed()
        availpkg = allpkg.available()
        availpkg = availpkg.latest()
        for pkg in instpkg:
            ipkgs.append(pkg.name)

        if lang not in self.langlist_with_cc:
            langcode = lang.split("_")[0]
        else:
            langcode = lang

        installed = self.check_if_already_installed(langcode)

        if installed is "no":
            all_pkg_list = self.get_language_packages(langcode)
            if all_pkg_list:
                pkgmatches = self.find_matching_pkgs(ipkgs, lang)

                # This is special case to cover package name man-pages-zh-CN
                # which should have been named as man-pages-zh_CN
                if lang.find("zh_CN") != -1:
                    pkgmatches.append("man-pages-zh-CN")

                # Available in repo pattern matched pkgs
                all_match_pkg_list = self.get_matches(availpkg, pkgmatches)

                # we want to make sure pkgs return only if
                # those pkgs are available to be installed
                for pkg in all_match_pkg_list:
                    if pkg not in ipkgs:
                        self.list_to_install.append(pkg)

                self.list_to_install.sort()
                print("Packages to be installed=> {0}".format(
                    self.list_to_install))

                if not self.list_to_install:
                    print("Langpack already installed")
                    self.add_langpack_to_installed_list(langcode)
                    return "LangPkgAlreadyInstalled"
                else:
                    ret = self.install_language_packages(langcode)
                    return ret
            else:
                print("We got no packages for requested language")
                return "LangPkgNoSupport"
        else:
            return "LangPackAlreadyInstalled"

    @classmethod
    def install_language_packages(cls, langcode):
        """ Install package list for a given langcode """
        cmd = "dnf langinstall -y {0}".format(langcode)
        try:
            subprocess.check_output(cmd, shell=True)
            return "LangSupportInstalled"
        except subprocess.CalledProcessError as e:
            print(e.message)

    def pk_install(self, langcode, pkglist_to_install):
        """ PK API to install required language support packages """
        bus_type = Gio.BusType.SESSION
        flags = 0
        iface_info = None

        # Creating proxies does not do any blocking I/O, and never fails
        proxy = Gio.DBusProxy.new_for_bus_sync(
            bus_type,
            flags,
            iface_info,
            'org.freedesktop.PackageKit',
            '/org/freedesktop/PackageKit',
            'org.freedesktop.PackageKit.Modify',
            None)
        try:
            proxy.set_default_timeout(0x7fffffff)  # timeout in milliseconds
            proxy.InstallPackageNames("(uass)", 0, pkglist_to_install,
                                      "show-confirm-search,hide-finished")
            self.add_langpack_to_installed_list(langcode)
            return "LangSupportInstalled"
        except GLib.GError as e:
            # When you click on cancel
            print("Unable to use PackageKit: {0}".format(e))
            return "LangPkgInstallCanceled"
